--[[
  possible game out of having certain shapes that make you lose?
  with harder levels having more (and more likely) shapes?
--]]

-- credit to http://lua-users.org/wiki/CopyTable for this method
function deepcopy(object)
  local lookup_table = {}
  local function _copy(object)
    if type(object) ~= "table" then
      return object
    elseif lookup_table[object] then
      return lookup_table[object]
    end
    local new_table = {}
    lookup_table[object] = new_table
    for index, value in pairs(object) do
      new_table[_copy(index)] = _copy(value)
    end
    return setmetatable(new_table, getmetatable(object))
  end
  return _copy(object)
end

function anyEqual(array, value)
  for _, thing in pairs(array) do
    if thing == value then return true end
  end
  return false
end

function love.load()
  -- constants
  grid_x = 8
  grid_y = 64
  pause_button = { 16, 32, 50, 20 }
  pause_wait = 0.3
  clear_button = { 116, 32, 50, 20 }
  clear_wait = 0.3
  pause_key = " "
  speed_up_key = "up"
  speed_down_key = "down"
  speed_up_one_key = "pageup"
  speed_down_one_key = "pagedown"
  speed_floor_key = "home"
  speed_ceil_key = "end"
  -- condition variables
  grid_width = 37
  grid_height = 33
  grid_size = 16
  birth_condition = { 3 }
  survive_condition = { 2, 3 }
  tick_length = 0.5
  -- game variables
  world = resetWorld()
  love.graphics.setBackgroundColor( 255, 255, 255, 255 )
  paused = true
  timer = 0
  pause_counter = 0
  clear_counter = 0
  timer_counter = 0
  tick_speed = 1
end

function resetWorld()
  local outcome = {}
  for i = 1, grid_width do
    outcome[i] = {}
    for j = 1,  grid_height do
      outcome[i][j] = false
    end
  end
  return outcome
end

function aliveNeighbours(x, y)
  local outcome = 0
  if world[x-1] then
    if world[x-1][y-1] then outcome = outcome + 1 end
    if world[x-1][y] then outcome = outcome + 1 end
    if world[x-1][y+1] then outcome = outcome + 1 end
  end
  if world[x] then
    if world[x][y-1] then outcome = outcome + 1 end
    -- don't count self!
    if world[x][y+1] then outcome = outcome + 1 end
  end
  if world[x+1] then
    if world[x+1][y-1] then outcome = outcome + 1 end
    if world[x+1][y] then outcome = outcome + 1 end
    if world[x+1][y+1] then outcome = outcome + 1 end
  end
  return outcome
end

function love.update(dt)
  -- update timers (timers are delays on button presses)
  pause_counter = pause_counter - dt
  clear_counter = clear_counter - dt
  timer_counter = timer_counter - dt
  -- pause/unpause if necessary
  if love.keyboard.isDown(pause_key) and pause_counter <= 0 then
    paused = not paused
    pause_counter = pause_wait
  end
  -- buttons!
  if love.mouse.isDown("l") then
    -- play/pause
    if (love.mouse.getX() > pause_button[1] and love.mouse.getX() < pause_button[3] + pause_button[1] and
        love.mouse.getY() > pause_button[2] and love.mouse.getY() < pause_button[4] + pause_button[2]) and pause_counter <= 0 then
      paused = not paused
      pause_counter = pause_wait
    -- reset world
    elseif (love.mouse.getX() > clear_button[1] and love.mouse.getX() < clear_button[3] + clear_button[1] and
            love.mouse.getY() > clear_button[2] and love.mouse.getY() < clear_button[4] + clear_button[2]) and clear_counter <= 0 then
      world = resetWorld()
      if not paused then pause = true end
      clear_counter = clear_wait
    end
  end
  if love.keyboard.isDown(speed_up_key) then
    tick_speed = tick_speed + 0.1
    if tick_speed > 60 then tick_speed = 60 end
  end if love.keyboard.isDown(speed_down_key) then
    tick_speed = tick_speed - 0.1
    if tick_speed < 0 then tick_speed = 0 end
  end
  if love.keyboard.isDown(speed_up_one_key) then
    tick_speed = tick_speed + 1
    if tick_speed > 60 then tick_speed = 60 end
  end if love.keyboard.isDown(speed_down_one_key) then
    tick_speed = tick_speed - 1
    if tick_speed < 0 then tick_speed = 0 end
  end
  if love.keyboard.isDown(speed_ceil_key) then
    tick_speed = math.ceil(tick_speed)
    if tick_speed > 60 then tick_speed = 60 end
  end if love.keyboard.isDown(speed_floor_key) then
    tick_speed = math.floor(tick_speed)
    if tick_speed < 0 then tick_speed = 0 end
  end
  -- if left-click create
  if love.mouse.isDown("l") then
    local x, y = love.mouse.getPosition()
    x = math.floor( (x  - grid_x) / grid_size )
    y = math.floor( (y  - grid_y) / grid_size )
    if x >= 1 and x <= grid_width and y >= 1 and y <= grid_height then
      world[x][y] = true
    end
  end
  -- if right-click destroy
  if love.mouse.isDown("r") then
    local x, y = love.mouse.getPosition()
    x = math.floor( (x  - grid_x) / grid_size )
    y = math.floor( (y  - grid_y) / grid_size )
    if x >= 1 and x <= grid_width and y >= 1 and y <= grid_height then
      world[x][y] = false
    end
  end
  -- if paused do no game updating
  if paused then 
    return 
  end
  -- update game timer
  timer = timer + (dt * tick_speed)
  -- if next tick
  if timer >= tick_length then
    -- copy world
    local world_copy = deepcopy(world)
    -- go through each world tile
    for i, column in pairs(world) do
      for j, cell in pairs(column) do
        -- get the count of its neighbours that are alive
        local neighbours = aliveNeighbours(i, j)
        -- if dead but birth conditions are met
        if not world[i][j] and anyEqual(birth_condition, neighbours) then
          -- birth!
          world_copy[i][j] = true
        end
        -- if alive and survive conditions are not met
        if world[i][j] and not anyEqual(survive_condition, neighbours) then
          -- die!
          world_copy[i][j] = false
        end
      end
    end
    world = world_copy
    timer = timer - tick_length
  end
end

function love.draw()
  -- draw grid
  love.graphics.setColor( 0, 0, 0, 255 )
  local x = (grid_width+1) / 2
  x = (x * grid_size)+ grid_x
  love.graphics.line(x, grid_y + grid_size, x + grid_size, grid_y + grid_size)
  local y = (grid_height+1) / 2
  y = (y * grid_size) + grid_y
  love.graphics.line(grid_x + grid_size, y, grid_x + grid_size, y + grid_size)
  love.graphics.setColor( 0, 0, 0, 64 )
  for i = 1, grid_width do
    for j = 1, grid_height do
      love.graphics.rectangle( "line", grid_x + (i * grid_size), grid_y + (j * grid_size), grid_size, grid_size )
    end
  end
  -- draw blocks
  love.graphics.setColor( 0, 0, 0, 255 )
  for i, column in pairs(world) do
    for j, cell in pairs(column) do
      if world[i][j] then
        love.graphics.rectangle( "fill", grid_x + (i * grid_size), grid_y + (j * grid_size), grid_size, grid_size )
      end
    end
  end
  -- draw buttons and "paused" if paused
  love.graphics.rectangle( "line", unpack(pause_button) )
  if paused then 
    love.graphics.print( "play", pause_button[1] + 10, pause_button[2] + 2 )
    love.graphics.print( "paused", 300, 60 )
  else 
    love.graphics.print( "pause", pause_button[1] + 5, pause_button[2] + 2 )
  end
  love.graphics.print( "speed: "..tick_speed, 300, 0 )
  love.graphics.rectangle( "line", unpack(clear_button) )
  love.graphics.print( "clear", clear_button[1] + 10, clear_button[2] + 2 )
  
end